;*********************************************************************
; DrumBox ~ MIDI to Drum Trigger Converter                           *
;*********************************************************************
; (c) 2004 SM-Sound / Seb Francis                                    *
;*********************************************************************
;
; Main code
;
;*********************************************************************
;
;    Connections:
;
;    RA0   <- Noise
;    RA1   -> MUX0
;    RA2   -> MUX1
;    RA3   -> MUX2
;    RA4   Open drain output!
;    RA5   -> MUX3
;
;    RB0   -> !TV1
;    RB1   -> !TV2
;    RB2   -> !TV3
;    RB3
;    RB4   <- !MIDI Channel bit0
;    RB5   <- !MIDI Channel bit1
;    RB6   <- !MIDI Channel bit2
;    RB7   <- !MIDI Channel bit3
;
;    RC0   Used by Timer1
;    RC1   Used by Timer1
;    RC2   -> DAC !CS
;    RC3   -> DAC SCLK
;    RC4   -> MUXE
;    RC5   -> DAC DIN
;    RC6
;    RC7   <- MIDI IN
;
;    RD0   -> !TV4
;    RD1   -> !TV5
;    RD2   -> !TV6
;    RD3   -> !TV7
;    RD4   -> !TV8
;    RD5   -> !TV9
;    RD6   -> !TV10
;    RD7   -> !TV11
;
;    RE0   -> !NV1
;    RE1   -> !NV2
;    RE2   -> !NV3
;
;
;*********************************************************************

; Compiler switches
 	#define	ENABLE_DELAYS

;*********************************************************************


	list      p=16f877	; List directive to define processor
	#include <p16f877.inc>
	#include <const.inc>	; Constants + Memory map
	#include <misc.inc>	; Various macros
	
	
	__CONFIG _CP_OFF & _WDT_OFF & _BODEN_ON & _PWRTE_ON & _HS_OSC & _WRT_ENABLE_OFF & _LVP_OFF & _DEBUG_OFF & _CPD_OFF 

; '__CONFIG' directive is used to embed configuration data within .asm file.
; The lables following the directive are located in the respective .inc file.
; See respective data sheet for additional information on configuration word.




;**********************************************************************


	org     0x0000		; Processor reset vector
	clrf    PCLATH		; Ensure page bits are cleared
	goto    Init		; Go to beginning of program


	org     0x0004		; Interrupt vector location
	movwf   W_TEMP		; Save W, STATUS, PCLATH and FSR
	swapf	STATUS,w	; Use swapf, not movf so not to disturb Z flag
	movwf	STATUS_TEMP
	movfw   PCLATH
	movwf	PCLATH_TEMP
	movfw	FSR
	movwf	FSR_TEMP
	
	clrf    PCLATH		; Clear page bits
	

; Call the interupt handlers

	clrf	STATUS		; RAM bank 0, IRP 0
	btfsc	INTCON,T0IF
	call	UpdateCV

	clrf	STATUS		; RAM bank 0, IRP 0
	btfsc	PIR1,RCIF
	call	MIDIByteReceived

; --
	movfw	FSR_TEMP	; Restore W, STATUS, PCLATH and FSR
	movwf	FSR
	movfw	PCLATH_TEMP
	movwf	PCLATH
	swapf	STATUS_TEMP,w	; Use swapf, not movf so not to disturb Z flag
	movwf	STATUS
	swapf	W_TEMP,f	; Use swapf, not movf so not to disturb Z flag
	swapf	W_TEMP,w	; Swap nibbles again to get them correct
	retfie			; Return from interrupt


;**********************************************************************


Init

;
; Setup the I/O and SFR
; 

; Bank 0 Setup
	bcf	STATUS,RP0
	bcf	STATUS,RP1
	
	clrf	PORTA		; MUX0, MUX1, MUX2, MUX3 = 0
	
	movlw	b'00000111'     ; !TV1->3 = 1
	movwf	PORTB
	
	clrf	PORTC           ; DAC !CS, DAC SCLK, MUXE, DAC DIN = 0
	
	movlw	b'11111111'     ; !TV4->11 = 1
	movwf	PORTD
	
	movlw	b'00000111'     ; !NV1->3 = 1
	movwf	PORTE
	
	clrf	PIR1		; Clear all interrupt flags
	clrf	PIR2
	
	movlw	b'00000000'	; A/D clock FOSC/2, Select RA0, A/D Off
	movwf	ADCON0

	movlw	b'00110000'	; TMR1 pre-scale 1:8, TMR1 osc disabled, Internal clock, TMR1 stopped
	movwf	T1CON
	
	movlw	b'10000000'	; USART: Enabled, 8bit reception, Reception (CREN) disabled, Address detection disabled
	movwf	RCSTA
	
	movlw	b'00000000'	; SPI: Disable serial port, Clock idle low, SPI mode @ 5MHz
	movwf	SSPCON
	
	; Initialise general data memory
	
	clrf	MIDI_FLAGS	; Clear MIDI flags
	clrf	MIDI_MSG1	; Clear MIDI status buffer
	
	movlw	low MIDI_MSG_BUFFER	; Set pointers to MIDI ring buffer
	movwf	MIDI_MSG_PROCESS_PTR
	movwf	MIDI_MSG_INCOMING_PTR
	
	
; Bank 1 Setup
	bsf	STATUS,RP0
	bcf	STATUS,RP1
	
	movlw	b'10001110'	; Right justified, RA0 analog, RA1-7 digital, Internal Vrefs
	movwf	ADCON1
	
	movlw	b'00000001'	; Port A (RA0 input)
	movwf	TRISA
	
	movlw	b'11110000'	; Port B (RB4-7 inputs)
	movwf	TRISB
	
	movlw	b'10000000'	; Port C (RC7 input)
	movwf	TRISC
	
	movlw	b'00000000'	; Port D (all outputs)
	movwf	TRISD
	
	movwf	TRISE		; Port E (all outputs)
	
	movlw	b'00000000'	; PORTB pullups enabled, Internal TMR0 clock, Pre-scaler assigned to TMR0, Pre-scaler ratio 1:2
	movwf	OPTION_REG	; TMR0 running for 333 instructions (166.5 ticks) gives update rate of 66.6us per channel, 999us overall
				; ISR with no other code (just updating CV output) takes 23us
				; S&H has 66.6-7.6 = 59us to settle (needs ~20us)
	
	movlw	b'01100000'	; Interupts: Global off, Peripheral on, TMR0 on, External off, RB off
	movwf	INTCON
	
	movlw	b'00100000'	; Peripheral interrupts 1: USART on, everything else off
	movwf	PIE1
	
	movlw	b'00000000'	; Peripheral interrupts 2: All off
	movwf	PIE2
	
	movlw	b'00000000'	; USART: Transmit disabled, Asynchronous, Low Baud Rate
	movwf	TXSTA
	
	movlw	D'9'		; USART: Baud rate 31,250
	movwf	SPBRG
	
	movlw	b'01000000'	; SPI: Transmit on rising edge
	movwf	SSPSTAT

	
	; Initialise general data memory

	movlw	1
	movwf	NEXT_CV		; Update CV1 next
	
	bcf	STATUS,IRP	; Indirectly address RAM banks 0-1
	
	; Set all CV levels to 0V (0x3F64)
	movlw	D'15'
	movwf	COUNTER
	movlw	low CV1_VALUE_L
	movwf	FSR
Init_CV_Loop
	movlw	0x64
	movwf	INDF
	incf	FSR
	movlw	0x3F
	movwf	INDF
	incf	FSR
	decfsz	COUNTER
	goto	Init_CV_Loop
	
	; Set all gate off counters to 0
	movlw	D'29'
	movwf	COUNTER
	movlw	low COUNT_V1
	movwf	FSR
Init_COUNT_Loop
	clrf	INDF
	incf	FSR
	decfsz	COUNTER
	goto	Init_COUNT_Loop
	
	; Set binary output states
	movlw	B'00000111'
	movwf	TV_VALUES_A
	movlw	B'11111111'
	movwf	TV_VALUES_B
	movlw	B'00000111'
	movwf	NV_VALUES


; Bank 2 Setup
; 	bcf	STATUS,RP0
; 	bsf	STATUS,RP1
	

; Bank 3 Setup
; 	bsf	STATUS,RP0
; 	bsf	STATUS,RP1
	
	bsf	STATUS,IRP	; Indirectly address RAM banks 2-3
	
	; Set pulse widths for all outputs to 1ms
	movlw	D'29'
	movwf	COUNTER
	movlw	low PW_V1
	movwf	FSR
Init_PW_Loop
	clrf	INDF
	bsf	INDF,0
	incf	FSR
	decfsz	COUNTER
	goto	Init_PW_Loop
	

; --------------------

	clrf	STATUS		; Bank 0 Select
	
	bsf	PORTC,2		; Set !CS high
	bsf	SSPCON,SSPEN	; SPI: Enable serial port
	bsf	RCSTA,CREN	; Enable MIDI receive
	bcf	INTCON,T0IF	; Reset TMR0 flag
	bsf	INTCON,GIE	; Enable interrupts


;#######################################################################

; RAM Bank 0 !
Main
	movfw	MIDI_MSG_PROCESS_PTR	; Check if buffer has any messages
	subwf	MIDI_MSG_INCOMING_PTR,w
	jz	Main
	
	movlw	high ProcessMIDI
	movwf	PCLATH			; Select program bank 2, and ready PCLATH for jump table
	call	ProcessMIDI
	clrf	PCLATH
	goto	Main
	

;#######################################################################



; --------------------------------


UpdateCV
	bsf	STATUS,RP1	; Select RAM bank 2
	
	movlw	D'91'		; Add to TMR0 to give 66.6us (999us overall) cycle time
	addwf	TMR0,f

	bsf	STATUS,RP0	; Select RAM bank 1
	bcf	STATUS,RP1
	
	; Jump to:
	; CV1 (0001)  -> 0x0800 (00001 000  0 0000000)
	; CV2 (0010)  -> 0x0880 (00001 000  1 0000000)
	; CV3 (0011)  -> 0x0900 (00001 001  0 0000000)
	; CV4 (0100)  -> 0x0980 (00001 001  1 0000000)
	; CV5 (0101)  -> 0x0A00 (00001 010  0 0000000)
	; CV6 (0110)  -> 0x0A80 (00001 010  1 0000000)
	; CV7 (0111)  -> 0x0B00 (00001 011  0 0000000)
	; CV8 (1000)  -> 0x0B80 (00001 011  1 0000000)
	; CV9 (1001)  -> 0x0C00 (00001 100  0 0000000)
	; CV10 (1010) -> 0x0C80 (00001 100  1 0000000)
	; CV11 (1011) -> 0x0D00 (00001 101  0 0000000)
	; CV12 (1100) -> 0x0D80 (00001 101  1 0000000)
	; CV13 (1101) -> 0x0E00 (00001 110  0 0000000)
	; CV14 (1110) -> 0x0E80 (00001 110  1 0000000)
	; CV15 (1111) -> 0x0F00 (00001 111  0 0000000)
	
	decf	NEXT_CV,w		; Make CV pointer 0 based
	movwf	PCLATH			; Set into PCLATH
	rrf	PCLATH			; Rotate right
	bsf	PCLATH,3		; Set bit 3
	clrf	TEMP1
	btfsc	STATUS,C		; Carry last bit into w
	bsf	TEMP1,7
	movfw	TEMP1
	movwf	PCL			; Jump to routine

; --------------------------------


PointToNextIncomingBufferByte	macro
	incf	FSR			; Increment pointer
	btfss	FSR,4			; Check not wrapped
	goto	$+3
	movlw	low MIDI_MSG_BUFFER	; Wrap pointer
	movwf	FSR
	
	movfw	FSR			; Check for ring buffer overrun
	subwf	MIDI_MSG_PROCESS_PTR,w
	jz	MIDIByteReceived_RingBuffOver
endm

MIDIByteReceived_RingBuffOver
	clrf	MIDI_MSG1		; Clear status byte and flags
	clrf	MIDI_FLAGS
	return


; RAM bank 0
MIDIByteReceived

	movfw	RCSTA			; Read and save and error code
	movwf	TEMP2
	
	movfw	RCREG			; Read MIDI byte
	movwf	TEMP1			; Save in TEMP1
	
	btfsc	TEMP2,OERR		; Check for overrun error
	goto	MIDIByteReceived_Error
	btfsc	TEMP2,FERR		; Check for framing error
	goto	MIDIByteReceived_Error
	
	
	btfss	TEMP1,7			; Header or data byte?
	goto	MIDIByteReceived_DataByte
	
; header byte
	movlw	b'11110000'		; System message
	andwf	TEMP1,w
	sublw	b'11110000'
	jz	MIDIByteReceived_System
	
	; Voice message
	clrf	MIDI_FLAGS		; Clear flags
	
	movfw	TEMP1			; Store status byte
	movwf	MIDI_MSG1
	
	movlw	b'11100000'		; 2 byte messages (Program Change + Channel Pressure)
	andwf	TEMP1,w
	sublw	b'11000000'
	sz				; If not 2 byte message, set flag
	bsf	MIDI_FLAGS,MSG_LEN
	
	return
	
MIDIByteReceived_System
	btfss	TEMP1,3			; System common message?
	goto	MIDIByteReceived_SystemCommon
	
	; System realtime message
	movfw	MIDI_MSG_INCOMING_PTR
	movwf	FSR
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	movfw	FSR
	movwf	MIDI_MSG_INCOMING_PTR
	
	movfw	TEMP1			; Write byte
	movwf	INDF
	
	return
	
MIDIByteReceived_SystemCommon
	clrf	MIDI_MSG1		; Clear status buffer
	
	return				; Ignore system common messages
	

MIDIByteReceived_DataByte
	movfw	MIDI_MSG1		; Drop byte if status is 0
	rz
	
	btfsc	MIDI_FLAGS,MSG_LEN	; Test if expecting 3 byte message
	goto	MIDIByteReceived_DataByte_3byte
	
	; 2 byte message
	movfw	MIDI_MSG_INCOMING_PTR
	movwf	FSR
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	
	movfw	MIDI_MSG1		; Write status byte
	movwf	INDF
	
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	movfw	FSR
	movwf	MIDI_MSG_INCOMING_PTR
	
	movfw	TEMP1			; Write data byte
	movwf	INDF
	
	return
	
MIDIByteReceived_DataByte_3byte
	btfsc	MIDI_FLAGS,WHICH_BYTE	; Test if this is the 2nd data byte
	goto	MIDIByteReceived_DataByte_3byte2
	
	; 3 byte message, 1st data byte
	movfw	TEMP1			; Store data byte
	movwf	MIDI_MSG2
	
	bsf	MIDI_FLAGS,WHICH_BYTE	; Set flag to expect 2nd data byte
	
	return
	
MIDIByteReceived_DataByte_3byte2
	; 3 byte message, 2nd data byte
	movfw	MIDI_MSG_INCOMING_PTR
	movwf	FSR
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	
	movfw	MIDI_MSG1		; Write status byte
	movwf	INDF
	
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	
	movfw	MIDI_MSG2		; Write 1st data byte
	movwf	INDF
	
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	movfw	FSR
	movwf	MIDI_MSG_INCOMING_PTR
	
	movfw	TEMP1			; Write 2nd data byte
	movwf	INDF
	
	bcf	MIDI_FLAGS,WHICH_BYTE	; Set flag to expect 1st data byte
	
	return
	
	
MIDIByteReceived_Error
	bcf	RCSTA,CREN		; Reset receive logic / clear error flag
	bsf	RCSTA,CREN
	return
	
	
	
;#######################################################################

; More include files can go here ..
	
;#######################################################################


; Program bank 1

	org	0x0800
; RAM bank 1
UpdateCV1
	call	UpdateCVFromRAM
	; RAM bank 1 still selected
	; STATUS,IRP now set for indirectly addressing RAM banks 0-1
	
	; Decrement pulse width counter if not already at 0
	movfw	COUNT_V1
	jz	UpdateCV1_BinOutputs
	decf	COUNT_V1
	jnz	UpdateCV1_BinOutputs
	; Counter just reached 0, so set DAC value for 0V
	movlw	0x3F
	movwf	CV1_VALUE_H
	movlw	0x64
	movwf	CV1_VALUE_L
	
UpdateCV1_BinOutputs
	; Set states of binary outputs TV1-TV3
	movfw	TV_VALUES_A
	bcf	STATUS,RP0	; Select RAM bank 0
	movwf	PORTB
	bsf	STATUS,RP0	; Select RAM bank 1
	
	; Decrement pulse width counters for binary outputs TV1-TV3
	movfw	COUNT_TV1
	jz	UpdateCV1_BinOutputs_TV2
	decf	COUNT_TV1
	snz
	; Counter just reached 0, so reset output state
	bsf	TV_VALUES_A,0
	
UpdateCV1_BinOutputs_TV2
	movfw	COUNT_TV2
	jz	UpdateCV1_BinOutputs_TV3
	decf	COUNT_TV2
	snz
	; Counter just reached 0, so reset output state
	bsf	TV_VALUES_A,1
	
UpdateCV1_BinOutputs_TV3
	movfw	COUNT_TV3
	jz	UpdateCV1_BinOutputs_End
	decf	COUNT_TV3
	snz
	; Counter just reached 0, so reset output state
	bsf	TV_VALUES_A,2
	
UpdateCV1_BinOutputs_End

	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0880
; RAM bank 1
UpdateCV2
	call	UpdateCVFromRAM
	; RAM bank 1 still selected
	; STATUS,IRP now set for indirectly addressing RAM banks 0-1
	
	; Decrement pulse width counter if not already at 0
	movfw	COUNT_V2
	jz	UpdateCV2_End
	decf	COUNT_V2
	jnz	UpdateCV2_End
	; Counter just reached 0, so set DAC value for 0V
	movlw	0x3F
	movwf	CV2_VALUE_H
	movlw	0x64
	movwf	CV2_VALUE_L
	
UpdateCV2_End
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0900
; RAM bank 1
UpdateCV3
	call	UpdateCVFromRAM
	; RAM bank 1 still selected
	; STATUS,IRP now set for indirectly addressing RAM banks 0-1
	
	; Decrement pulse width counter if not already at 0
	movfw	COUNT_V3
	jz	UpdateCV3_End
	decf	COUNT_V3
	jnz	UpdateCV3_End
	; Counter just reached 0, so set DAC value for 0V
	movlw	0x3F
	movwf	CV3_VALUE_H
	movlw	0x64
	movwf	CV3_VALUE_L
	
UpdateCV3_End
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0980
; RAM bank 1
UpdateCV4
	call	UpdateCVFromRAM
	; RAM bank 1 still selected
	; STATUS,IRP now set for indirectly addressing RAM banks 0-1
	
	; Decrement pulse width counter if not already at 0
	movfw	COUNT_V4
	jz	UpdateCV4_End
	decf	COUNT_V4
	jnz	UpdateCV4_End
	; Counter just reached 0, so set DAC value for 0V
	movlw	0x3F
	movwf	CV4_VALUE_H
	movlw	0x64
	movwf	CV4_VALUE_L
	
UpdateCV4_End
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0A00
; RAM bank 1
UpdateCV5
	call	UpdateCVFromRAM
	; RAM bank 1 still selected
	; STATUS,IRP now set for indirectly addressing RAM banks 0-1
	
	; Decrement pulse width counter if not already at 0
	movfw	COUNT_V5
	jz	UpdateCV5_End
	decf	COUNT_V5
	jnz	UpdateCV5_End
	; Counter just reached 0, so set DAC value for 0V
	movlw	0x3F
	movwf	CV5_VALUE_H
	movlw	0x64
	movwf	CV5_VALUE_L
	
UpdateCV5_End
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0A80
; RAM bank 1
UpdateCV6
	call	UpdateCVFromRAM
	; RAM bank 1 still selected
	; STATUS,IRP now set for indirectly addressing RAM banks 0-1
	
	; Decrement pulse width counter if not already at 0
	movfw	COUNT_V6
	jz	UpdateCV6_BinOutputs
	decf	COUNT_V6
	jnz	UpdateCV6_BinOutputs
	; Counter just reached 0, so set DAC value for 0V
	movlw	0x3F
	movwf	CV6_VALUE_H
	movlw	0x64
	movwf	CV6_VALUE_L
	
UpdateCV6_BinOutputs
	; Set states of binary outputs TV4-TV11
	movfw	TV_VALUES_B
	bcf	STATUS,RP0	; Select RAM bank 0
	movwf	PORTD
	bsf	STATUS,RP0	; Select RAM bank 1
	
	; Decrement pulse width counters for binary outputs TV4-TV11
	movfw	COUNT_TV4
	jz	UpdateCV6_BinOutputs_TV5
	decf	COUNT_TV4
	snz
	; Counter just reached 0, so reset output state
	bsf	TV_VALUES_B,0
	
UpdateCV6_BinOutputs_TV5
	movfw	COUNT_TV5
	jz	UpdateCV6_BinOutputs_TV6
	decf	COUNT_TV5
	snz
	; Counter just reached 0, so reset output state
	bsf	TV_VALUES_B,1
	
UpdateCV6_BinOutputs_TV6
	movfw	COUNT_TV6
	jz	UpdateCV6_BinOutputs_TV7
	decf	COUNT_TV6
	snz
	; Counter just reached 0, so reset output state
	bsf	TV_VALUES_B,2
	
UpdateCV6_BinOutputs_TV7
	movfw	COUNT_TV7
	jz	UpdateCV6_BinOutputs_TV8
	decf	COUNT_TV7
	snz
	; Counter just reached 0, so reset output state
	bsf	TV_VALUES_B,3
	
UpdateCV6_BinOutputs_TV8
	movfw	COUNT_TV8
	jz	UpdateCV6_BinOutputs_TV9
	decf	COUNT_TV8
	snz
	; Counter just reached 0, so reset output state
	bsf	TV_VALUES_B,4
	
UpdateCV6_BinOutputs_TV9
	movfw	COUNT_TV9
	jz	UpdateCV6_BinOutputs_TV10
	decf	COUNT_TV9
	snz
	; Counter just reached 0, so reset output state
	bsf	TV_VALUES_B,5
	
UpdateCV6_BinOutputs_TV10
	movfw	COUNT_TV10
	jz	UpdateCV6_BinOutputs_TV11
	decf	COUNT_TV10
	snz
	; Counter just reached 0, so reset output state
	bsf	TV_VALUES_B,6
	
UpdateCV6_BinOutputs_TV11
	movfw	COUNT_TV11
	jz	UpdateCV6_BinOutputs_End
	decf	COUNT_TV11
	snz
	; Counter just reached 0, so reset output state
	bsf	TV_VALUES_B,7
	
UpdateCV6_BinOutputs_End

	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0B00
; RAM bank 1
UpdateCV7
	call	UpdateCVFromRAM
	; RAM bank 1 still selected
	; STATUS,IRP now set for indirectly addressing RAM banks 0-1
	
	; Decrement pulse width counter if not already at 0
	movfw	COUNT_V7
	jz	UpdateCV7_End
	decf	COUNT_V7
	jnz	UpdateCV7_End
	; Counter just reached 0, so set DAC value for 0V
	movlw	0x3F
	movwf	CV7_VALUE_H
	movlw	0x64
	movwf	CV7_VALUE_L
	
UpdateCV7_End
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0B80
; RAM bank 1
UpdateCV8
	call	UpdateCVFromRAM
	; RAM bank 1 still selected
	; STATUS,IRP now set for indirectly addressing RAM banks 0-1
	
	; Decrement pulse width counter if not already at 0
	movfw	COUNT_V8
	jz	UpdateCV8_End
	decf	COUNT_V8
	jnz	UpdateCV8_End
	; Counter just reached 0, so set DAC value for 0V
	movlw	0x3F
	movwf	CV8_VALUE_H
	movlw	0x64
	movwf	CV8_VALUE_L
	
UpdateCV8_End
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0C00
; RAM bank 1
UpdateCV9
	call	UpdateCVFromRAM
	; RAM bank 1 still selected
	; STATUS,IRP now set for indirectly addressing RAM banks 0-1
	
	; Decrement pulse width counter if not already at 0
	movfw	COUNT_V9
	jz	UpdateCV9_End
	decf	COUNT_V9
	jnz	UpdateCV9_End
	; Counter just reached 0, so set DAC value for 0V
	movlw	0x3F
	movwf	CV9_VALUE_H
	movlw	0x64
	movwf	CV9_VALUE_L
	
UpdateCV9_End
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0C80
; RAM bank 1
UpdateCV10
	call	UpdateCVFromRAM
	; RAM bank 1 still selected
	; STATUS,IRP now set for indirectly addressing RAM banks 0-1
	
	; Decrement pulse width counter if not already at 0
	movfw	COUNT_V10
	jz	UpdateCV10_End
	decf	COUNT_V10
	jnz	UpdateCV10_End
	; Counter just reached 0, so set DAC value for 0V
	movlw	0x3F
	movwf	CV10_VALUE_H
	movlw	0x64
	movwf	CV10_VALUE_L
	
UpdateCV10_End
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0D00
; RAM bank 1
UpdateCV11
	call	UpdateCVFromRAM
	; RAM bank 1 still selected
	; STATUS,IRP now set for indirectly addressing RAM banks 0-1
	
	; Decrement pulse width counter if not already at 0
	movfw	COUNT_V11
	jz	UpdateCV11_BinOutputs
	decf	COUNT_V11
	jnz	UpdateCV11_BinOutputs
	; Counter just reached 0, so set DAC value for 0V
	movlw	0x3F
	movwf	CV11_VALUE_H
	movlw	0x64
	movwf	CV11_VALUE_L
	
UpdateCV11_BinOutputs
	; Set states of binary outputs NV1-NV3
	movfw	NV_VALUES
	bcf	STATUS,RP0	; Select RAM bank 0
	movwf	PORTE
	bsf	STATUS,RP0	; Select RAM bank 1
	
	; Decrement pulse width counters for binary outputs NV1-NV3
	movfw	COUNT_NV1
	jz	UpdateCV11_BinOutputs_NV2
	decf	COUNT_NV1
	snz
	; Counter just reached 0, so reset output state
	bsf	NV_VALUES,0
	
UpdateCV11_BinOutputs_NV2
	movfw	COUNT_NV2
	jz	UpdateCV11_BinOutputs_NV3
	decf	COUNT_NV2
	snz
	; Counter just reached 0, so reset output state
	bsf	NV_VALUES,1
	
UpdateCV11_BinOutputs_NV3
	movfw	COUNT_NV3
	jz	UpdateCV11_BinOutputs_End
	decf	COUNT_NV3
	snz
	; Counter just reached 0, so reset output state
	bsf	NV_VALUES,2
	
UpdateCV11_BinOutputs_End

	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0D80
; RAM bank 1
UpdateCV12
	call	UpdateCVFromRAM
	; RAM bank 1 still selected
	; STATUS,IRP now set for indirectly addressing RAM banks 0-1
	
	; Decrement pulse width counter if not already at 0
	movfw	COUNT_V12
	jz	UpdateCV12_End
	decf	COUNT_V12
	jnz	UpdateCV12_End
	; Counter just reached 0, so set DAC value for 0V
	movlw	0x3F
	movwf	CV12_VALUE_H
	movlw	0x64
	movwf	CV12_VALUE_L
	
UpdateCV12_End
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0E00
; RAM bank 1
UpdateCV13
	call	UpdateCVFromRAM
	; RAM bank 1 still selected
	; STATUS,IRP now set for indirectly addressing RAM banks 0-1
	
	; Decrement pulse width counter if not already at 0
	movfw	COUNT_V13
	jz	UpdateCV13_End
	decf	COUNT_V13
	jnz	UpdateCV13_End
	; Counter just reached 0, so set DAC value for 0V
	movlw	0x3F
	movwf	CV13_VALUE_H
	movlw	0x64
	movwf	CV13_VALUE_L
	
UpdateCV13_End
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0E80
; RAM bank 1
UpdateCV14
	call	UpdateCVFromRAM
	; RAM bank 1 still selected
	; STATUS,IRP now set for indirectly addressing RAM banks 0-1
	
	; Decrement pulse width counter if not already at 0
	movfw	COUNT_V14
	jz	UpdateCV14_End
	decf	COUNT_V14
	jnz	UpdateCV14_End
	; Counter just reached 0, so set DAC value for 0V
	movlw	0x3F
	movwf	CV14_VALUE_H
	movlw	0x64
	movwf	CV14_VALUE_L
	
UpdateCV14_End
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0F00
; RAM bank 1
UpdateCV15
	call UpdateCVFromRAM
	; RAM bank 1 still selected
	; STATUS,IRP now set for indirectly addressing RAM banks 0-1
	
	; Decrement pulse width counter if not already at 0
	movfw	COUNT_V15
	jz	UpdateCV15_End
	decf	COUNT_V15
	jnz	UpdateCV15_End
	; Counter just reached 0, so set DAC value for 0V
	movlw	0x3F
	movwf	CV15_VALUE_H
	movlw	0x64
	movwf	CV15_VALUE_L
	
UpdateCV15_End
	clrf	NEXT_CV		; Point to CV1
	incf	NEXT_CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------

; RAM bank 1
UpdateCVFromRAM
	bcf	STATUS,IRP	; Indirectly address RAM bank 0-1
	
	movlw	low (CV1_VALUE_L - 2)	; Base address
	movwf	FSR
	movfw	NEXT_CV		; Add 2xoffset to base address
	addwf	FSR,f
	addwf	FSR,f
	
	movwf	TEMP3		; Save offset in bank 0 accessible RAM
	
	bcf	STATUS,RP0	; Select RAM bank 0
	
	movfw	INDF		; Get LSB
	movwf	TEMP1
	incf	FSR		; Get MSB
	movfw	INDF
	movwf	TEMP2
	
	addlw	0		; Clear carry flag
	rlf	TEMP1		; Shift bits 2 places to left
	rlf	TEMP2
	rlf	TEMP1
	rlf	TEMP2,w
	
	; Set DAC to voltage
	bcf	PORTC,DACCS	; Set !CS low
	movwf	SSPBUF		; Transmit MSB
	nop2			; Wait 9 cycles before sending next byte
	nop2
	nop2
	nop
	bcf	MUXE		; While waiting: Turn off MUX (takes max 300ns)  ; Maybe move this to earlier to eliminate digital feedthrough from the SPI signal???
	movfw	TEMP1		; While waiting: Load LSB into w
	movwf	SSPBUF		; Transmit LSB and sub-bits
	nop2			; Wait 8 cycles
	rlf	TEMP3		; While waiting: Shift MUX bits to left
	btfsc	TEMP3,4		; While waiting: Copy bit 4 to left
	bsf	TEMP3,5
	movfw	TEMP3
	movwf	PORTA		; While waiting: Select MUX channel
	nop
	bsf	PORTC,DACCS	; Set !CS high (loads DAC)
	
	wait_us 5		; Wait 5us for DAC and opamps to settle
				; Maybe less wait with faster opamp?
	
	bsf	MUXE		; Turn on MUX (takes typ 105ns, max 600ns)
	
	; S&H now has until next call of UpdateCVFromRAM to settle

	bsf	STATUS,RP0	; Select RAM bank 1
	
	return


;#######################################################################


	; Program bank 2
	
	org	0x1000
	
	; Implemented messages:
	;
	; 1001cccc nnnnnnnn vvvvvvvv  Note On
	; 1010cccc nnnnnnnn pppppppp  Polyphonic Aftertouch


PointToNextProcessBufferByte	macro
	incf	MIDI_MSG_PROCESS_PTR	; Increment pointer
	btfss	MIDI_MSG_PROCESS_PTR,4	; Check not wrapped
	goto	$+3
	movlw	low MIDI_MSG_BUFFER	; Wrap pointer
	movwf	MIDI_MSG_PROCESS_PTR
endm


; Parse MIDI message here

ProcessMIDI
; RAM Bank 0
	PointToNextProcessBufferByte
	
	bcf	STATUS,IRP	; Indirectly address banks 0-1
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	
	movfw	INDF		; Get and save MIDI channel
	andlw	b'00001111'
	movwf	MTEMP1
	
	rrf	INDF,w		; Get message
	movwf	MTEMP2
	rrf	MTEMP2
	rrf	MTEMP2
	rrf	MTEMP2,w
	andlw	b'00000111'
	
	addwf	PCL		; Jump to handler (PCLATH already set before call to ProcessMIDI)
	goto	ProcessMIDI_NoteOff
	goto	ProcessMIDI_NoteOn
	goto	ProcessMIDI_PolyAftertouch
	goto	ProcessMIDI_ControlChange
	goto	ProcessMIDI_ProgramChange
	goto	ProcessMIDI_Aftertouch
	goto	ProcessMIDI_PitchBend
	goto	ProcessMIDI_System


;***********************************************************************
;***********************************************************************
	
; Message handlers

ProcessMIDI_NoteOff
	PointToNextProcessBufferByte
	PointToNextProcessBufferByte
	return
	
	
;***********************************************************************
	
	
ProcessMIDI_NoteOn
; RAM Bank 0
	PointToNextProcessBufferByte
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	
	; Check MIDI channel
	rrf	PORTB,w
	movwf	MTEMP3
	rrf	MTEMP3
	rrf	MTEMP3
	rrf	MTEMP3
	comf	MTEMP3,w
	andlw	B'00001111'
	subwf	MTEMP1,w
	jnz	ProcessMIDI_NoteOn_End
	
	; Lookup which output note value maps to
	GVFT	TriggerOutput,INDF
	movwf	MTEMP1
	
	; If 0x00 (not used) do nothing
	movfw	MTEMP1
	jz	ProcessMIDI_NoteOn_End
	
	; Output type to MTEMP2
	andlw	B'11110000'
	movwf	MTEMP2
	; Output number to MTEMP1
	movlw	B'00001111'
	andwf	MTEMP1,f
	
	; Get note velocity to MTEMP3
	PointToNextProcessBufferByte
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	movfw	INDF
	rz			; Return if velocity = 0
	movwf	MTEMP3
	
	; Switch to code to handle each output type
	movlw	0x10		; Variable pulse amplitude
	subwf	MTEMP2,w
	jz	GateOn_V
	movlw	0x20		; Trimmer-set pulse amplitude
	subwf	MTEMP2,w
	jz	GateOn_TV
				; Else fixed pulse amplitude
				
				
GateOn_NV
	bsf	STATUS,RP0	; Bank 1 Select
	
	; Jump to appropriate output set code
	movlw	high (GateOn_NV_SetOutput - 2)
	movwf	PCLATH
	movlw	low (GateOn_NV_SetOutput - 2)
	addwf	MTEMP1,w
	snc
	incf	PCLATH
	addwf	MTEMP1,w
	snc
	incf	PCLATH
	bcf	INTCON,GIE	; Disable interrupts before setting output state
	movwf	PCL
	
	; Set output state
GateOn_NV_SetOutput
	bcf	NV_VALUES,0
	goto	GateOn_NV_OutputSet
	bcf	NV_VALUES,1
	goto	GateOn_NV_OutputSet
	bcf	NV_VALUES,2
	
GateOn_NV_OutputSet
	bcf	STATUS,RP0	; Bank 0 Select
	
; Set counter for pulse width
	; Get current pulse width for this output to MTEMP3
	bsf	STATUS,IRP	; Indirectly address banks 2-3
	movlw	low (PW_NV1 - 1)
	movwf	FSR
	movfw	MTEMP1
	addwf	FSR,f
	movfw	INDF
	movwf	MTEMP3
	
	; Write MTEMP3 to counter register
	bcf	STATUS,IRP	; Indirectly address banks 0-1
	movlw	low (COUNT_NV1 - 1)
	movwf	FSR
	movfw	MTEMP1
	addwf	FSR,f
	movfw	MTEMP3
	movwf	INDF
	
	; Enable interrupts again
	bsf	INTCON,GIE
	
	return


GateOn_TV
	bsf	STATUS,RP0	; Bank 1 Select
	
	; Jump to appropriate output set code
	movlw	high (GateOn_TV_SetOutput - 2)
	movwf	PCLATH
	movlw	low (GateOn_TV_SetOutput - 2)
	addwf	MTEMP1,w
	snc
	incf	PCLATH
	addwf	MTEMP1,w
	snc
	incf	PCLATH
	bcf	INTCON,GIE	; Disable interrupts before setting output state
	movwf	PCL
	
	; Set output state
GateOn_TV_SetOutput
	bcf	TV_VALUES_A,0
	goto	GateOn_TV_OutputSet
	bcf	TV_VALUES_A,1
	goto	GateOn_TV_OutputSet
	bcf	TV_VALUES_A,2
	goto	GateOn_TV_OutputSet
	bcf	TV_VALUES_B,0
	goto	GateOn_TV_OutputSet
	bcf	TV_VALUES_B,1
	goto	GateOn_TV_OutputSet
	bcf	TV_VALUES_B,2
	goto	GateOn_TV_OutputSet
	bcf	TV_VALUES_B,3
	goto	GateOn_TV_OutputSet
	bcf	TV_VALUES_B,4
	goto	GateOn_TV_OutputSet
	bcf	TV_VALUES_B,5
	goto	GateOn_TV_OutputSet
	bcf	TV_VALUES_B,6
	goto	GateOn_TV_OutputSet
	bcf	TV_VALUES_B,7
	
GateOn_TV_OutputSet
	bcf	STATUS,RP0	; Bank 0 Select
	
; Set counter for pulse width
	; Get current pulse width for this output to MTEMP3
	bsf	STATUS,IRP	; Indirectly address banks 2-3
	movlw	low (PW_TV1 - 1)
	movwf	FSR
	movfw	MTEMP1
	addwf	FSR,f
	movfw	INDF
	movwf	MTEMP3
	
	; Write MTEMP3 to counter register
	bcf	STATUS,IRP	; Indirectly address banks 0-1
	movlw	low (COUNT_TV1 - 1)
	movwf	FSR
	movfw	MTEMP1
	addwf	FSR,f
	movfw	MTEMP3
	movwf	INDF
	
	; Enable interrupts again
	bsf	INTCON,GIE
	
	return


GateOn_V
	; Load note velocity to W
	movfw	MTEMP3
	
	; Calculate DAC value for this velocity
	sublw	D'127'		; (127 - Velocity)
	movwf	AARGB0		; * 0x5C
	movlw	0x5C
	movwf	BARGB0
	call	FXM0707U
	
	; Point to CV register
	movlw	low (CV1_VALUE_L - 2)
	movwf	FSR
	movfw	MTEMP1
	addwf	FSR,f
	addwf	FSR,f
	
	; Disable interrupts
	bcf	INTCON,GIE
	
	; Write pulse amplitude DAC value
	movfw	AARGB1
	movwf	INDF
	movfw	AARGB0
	incf	FSR
	movwf	INDF
	
; Set counter for pulse width
	; Get current pulse width for this output to MTEMP3
	bsf	STATUS,IRP	; Indirectly address banks 2-3
	movlw	low (PW_V1 - 1)
	movwf	FSR
	movfw	MTEMP1
	addwf	FSR,f
	movfw	INDF
	movwf	MTEMP3
	
	; Write MTEMP3 to counter register
	bcf	STATUS,IRP	; Indirectly address banks 0-1
	movlw	low (COUNT_V1 - 1)
	movwf	FSR
	movfw	MTEMP1
	addwf	FSR,f
	movfw	MTEMP3
	movwf	INDF
	
	; Enable interrupts again
	bsf	INTCON,GIE
	
	return
	
ProcessMIDI_NoteOn_End
	PointToNextProcessBufferByte
	return
	
	
;***********************************************************************
	
	
ProcessMIDI_PolyAftertouch
; RAM Bank 0
	PointToNextProcessBufferByte
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	
	; Check MIDI channel
	rrf	PORTB,w
	movwf	MTEMP3
	rrf	MTEMP3
	rrf	MTEMP3
	rrf	MTEMP3
	comf	MTEMP3,w
	andlw	B'00001111'
	subwf	MTEMP1,w
	jnz	ProcessMIDI_PolyAftertouch_End
	
	; Lookup which output note value maps to
	GVFT	TriggerOutput,INDF
	movwf	MTEMP1
	
	; If 0x00 (not used) do nothing
	movfw	MTEMP1
	jz	ProcessMIDI_PolyAftertouch_End
	
	; Output type to MTEMP2
	andlw	B'11110000'
	movwf	MTEMP2
	; Output number to MTEMP1
	movlw	B'00001111'
	andwf	MTEMP1,f
	
	; Get note aftertouch value to MTEMP3
	PointToNextProcessBufferByte
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	movfw	INDF
	snz			; If 0, set to 1
	movlw	D'1'
	movwf	MTEMP3
	
	; Switch to code to handle each output type
	movlw	0x10		; Variable pulse amplitude
	subwf	MTEMP2,w
	jz	SetPulseWidth_V
	movlw	0x20		; Trimmer-set pulse amplitude
	subwf	MTEMP2,w
	jz	SetPulseWidth_TV
				; Else fixed pulse amplitude
SetPulseWidth_NV
	; Set pulse width for output to aftertouch value (MTEMP3)
	bsf	STATUS,IRP	; Indirectly address banks 2-3
	movlw	low (PW_NV1 - 1)
	movwf	FSR
	movfw	MTEMP1
	addwf	FSR,f
	movfw	MTEMP3
	movwf	INDF
	return
	
SetPulseWidth_TV
	; Set pulse width for output to aftertouch value (MTEMP3)
	bsf	STATUS,IRP	; Indirectly address banks 2-3
	movlw	low (PW_TV1 - 1)
	movwf	FSR
	movfw	MTEMP1
	addwf	FSR,f
	movfw	MTEMP3
	movwf	INDF
	return

SetPulseWidth_V
	; Set pulse width for output to aftertouch value (MTEMP3)
	bsf	STATUS,IRP	; Indirectly address banks 2-3
	movlw	low (PW_V1 - 1)
	movwf	FSR
	movfw	MTEMP1
	addwf	FSR,f
	movfw	MTEMP3
	movwf	INDF
	return

ProcessMIDI_PolyAftertouch_End
	PointToNextProcessBufferByte
	return
	
	
;***********************************************************************
	
	
ProcessMIDI_ControlChange
	PointToNextProcessBufferByte
	PointToNextProcessBufferByte
	return
	
	
;***********************************************************************
	
	
ProcessMIDI_ProgramChange
	PointToNextProcessBufferByte
	return
	
	
;***********************************************************************
	
	
ProcessMIDI_Aftertouch
	PointToNextProcessBufferByte
	return
	
	
;***********************************************************************
	
	
ProcessMIDI_PitchBend
	PointToNextProcessBufferByte
	PointToNextProcessBufferByte
	return
	
	
;***********************************************************************
	
	
ProcessMIDI_System
	return
	
	
;***********************************************************************
;***********************************************************************
	
	
; Note value to trigger output maps
#include <notemap.inc>

; Maths Functions
#include <maths.inc>



;#######################################################################


	; Program bank 3
	
	org	0x1800
	nop
	

;#######################################################################


	END                       ; directive 'end of program'


